import argparse
import os
import sys
from subprocess import run
from typing import Dict, Tuple, List

from yaml import safe_load

from mungo.common import vparse, print_dag, mungoversion
from mungo.repositorydata import RepositoryData
from mungo.solver import create_dag, nodes_to_install


def _install_prompt(to_install: list, installed: dict = None, ask_for_confirmation: bool = True) -> bool:
    if len(to_install) == 0:
        print("Nothing to do.")
        exit(0)

    def _prepare_packages(packages: dict) -> dict:
        # update package information to contain everything we need for displaying relevant information
        for package, ds in packages.items():
            for data in ds:
                if data is not None:
                    data['version_build'] = data['version'] + '-' + data['build']
                    if "channel" not in data:
                        data['reponame'] = "local"  # TODO: replace this
                    else:
                        data['reponame'] = data['channel'].split('/')[-2]
        return packages

    def _calculate_widths(packages: dict) -> Dict[str, int]:
        # calculate column-widths for the three columns 'name', 'version_build' and 'reponame'
        widths = {t: max([max([len(d_from[t]) if d_from is not None else 0, len(d_to[t])])
                          for (d_from, d_to) in packages.values()]) if len(packages) > 0 else 0
                  for t in ['name', 'version_build', 'reponame']}
        return widths

    def _print_packages(packages: dict, widths: Dict[str, int]):
        # print packages that are to be installed / updated / downgraded
        print("")
        for (package_from, package_to) in packages.values():
            if "installed" not in package_to:  # only display packages that aren't already installed
                if package_from is not None:  # if there's an update/downgrade of a package
                    print("\t{:<{name_width}} {:<{version_width}} {:<{repo_width}}  ->  "
                          "{:<{version_width}} {:<{repo_width}}"
                          .format(package_from['name'], package_from['version_build'], package_from['reponame'],
                                  package_to['version_build'], package_to['reponame'],
                                  name_width=widths['name'],
                                  version_width=widths['version_build'],
                                  repo_width=widths['reponame']))
                else:  # install a new package (not an update/downgrade of an existing package)
                    print("\t{:<{name_width}} {:<{version_width}} {:<{repo_width}}"
                          .format(package_to['name'], package_to['version_build'], package_to['reponame'],
                                  name_width=widths['name'],
                                  version_width=widths['version_build'],
                                  repo_width=widths['reponame']))
        print("")

    to_install = {node.p['name']: node.p for node in to_install}

    if installed is not None:
        # which packages have changed?
        changed = installed.keys() & to_install.keys()

        # assert that there is only one version installed.
        # TODO: What do we do if there are multiple versions?
        if not all(len(installed[package].keys()) == 1 for package in installed.keys()):
            print(f"WARNING: Multiple locally installed versions for the same package.", file=sys.stderr)

        local_versions = {name: list(installed[name].values())[0] for name in changed}
        remote_versions = {name: data for name, data in to_install.items() if name in changed}
        version_changes = {package: (local_versions[package], remote_versions[package]) for package in changed}

        upgrades = {package: (p_from, p_to)
                    for (package, (p_from, p_to)) in version_changes.items()
                    if vparse(p_from['version']) < vparse(p_to['version'])
                    }
        # if vparse(([p_from['version'], p_from['build']])) < vparse(([p_to['version'], p_to['build']]))
        downgrades = {package: (p_from, p_to)
                      for (package, (p_from, p_to)) in version_changes.items()
                      # if vparse(([p_from['version'], p_from['build']])) > vparse(([p_to['version'], p_to['build']]))}
                      if vparse(p_from['version']) > vparse(p_to['version'])
                      }

        # remove packages that are to be updated or downgraded from the list of *new* packages
        for package in upgrades.keys() | downgrades.keys():
            to_install.pop(package, None)
    else:
        upgrades = dict()
        downgrades = dict()

    # add a dummy "package_from" (`None`) to get rid of special cases
    to_install = {package: (None, p_to) for (package, p_to) in to_install.items()}

    # get maximum column widths per type
    # across all three modes of package installation (install/upgrade/downgrade)
    all_widths = [_calculate_widths(_prepare_packages(packages)) for packages in [to_install, upgrades, downgrades]]
    widths = dict()
    for w in all_widths:
        for k, v in w.items():
            widths[k] = max(widths.get(k, 0), v)

    if len(upgrades) > 0:
        print("The following packages will be UPDATED:")
        _print_packages(upgrades, widths)

    if len(downgrades) > 0:
        print("The following packages will be DOWNGRADED:")
        _print_packages(downgrades, widths)

    if len(to_install) > 0:
        print("The following packages will be INSTALLED:")
        _print_packages(to_install, widths)

    if ask_for_confirmation:
        answer = input("\nProceed ([y]/n)? ")
        if answer.lower() not in {'\n', 'y', ''}:
            return False
    return True


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("-j", "--jobs", type=int, help="number of jobs", default=8)
    parser.add_argument("--debug", action="store_true")
    parser.add_argument('--version', action='version', version=f'mungo {mungoversion}')

    subparsers = parser.add_subparsers(title='subcommands',
                                       dest='command',
                                       description='valid subcommands',
                                       help='additional help')
    subparsers.required = True

    parser_create = subparsers.add_parser('create')
    parser_create.add_argument("-n", "--name", help="the name of the environment")
    parser_create.add_argument("package_spec", nargs="*")
    parser_create.add_argument("--file", metavar="FILE", default=None)
    parser_create.add_argument("--channel", "-c", nargs=1, action="append", metavar="CHANNEL", default=[])
    parser_create.add_argument("--dag", action="store_true", default=False, help="Do not execute anything and print "
                                                                                 "the directed acyclic graph of "
                                                                                 "package dependency in the dot "
                                                                                 "language. Recommended use on Unix "
                                                                                 "systems: mungo create --dag | dot | "
                                                                                 "display")
    parser_create.add_argument("-y", "--yes", help="Do not ask for confirmation.", action="store_true")
    parser_create.add_argument("--offline", action="store_true")
    parser_create.add_argument("--force_download", action="store_true")

    parser_install = subparsers.add_parser('install')
    parser_install.add_argument("-n", "--name", help="the name of the environment", default=None)
    parser_install.add_argument("package_spec", nargs="+")
    parser_install.add_argument("--file", metavar="FILE", default=None)
    parser_install.add_argument("--channel", "-c", nargs=1, action="append", metavar="CHANNEL", default=[])
    parser_install.add_argument("--dag", action="store_true", default=False, help="Do not execute anything and print "
                                                                                  "the directed acyclic graph of "
                                                                                  "package dependency in the dot "
                                                                                  "language. Recommended use on Unix "
                                                                                  "systems: mungo create --dag | dot "
                                                                                  "| display")
    parser_install.add_argument("-y", "--yes", help="Do not ask for confirmation.", action="store_true")
    parser_install.add_argument("--offline", action="store_true")
    parser_install.add_argument("--force_download", action="store_true")

    args = parser.parse_args()
    # use either the supplied name, the current prefix or the default env 'base' in that order
    args.name = args.name or os.getenv('CONDA_DEFAULT_ENV', 'base')
    jobs = args.jobs
    command = args.command
    dag = args.dag
    ask_for_confirmation = not args.yes
    offline = args.offline
    force_download = args.force_download
    debug = args.debug

    channels = [channel for channel in args.channel]
    channels = [l for channel in channels for l in channel if isinstance(channel, list)]  # flatten channel lists
    packages = set(args.package_spec)
    if args.file is not None:
        name, channels_rc, packages_rc, pip = _read_environment_description(args.file)
        name = name or args.name  # if the environment description does not supply a name, use args.name instead
        channels.extend(channels_rc)
        packages |= set(packages_rc)
    else:
        name = args.name
        pip = []

    default_channels = _get_condarc_channels()
    channels.extend(default_channels)

    print("WARNING: mungo is experimental, use at your own risk.", file=sys.stderr)
    if command == "create":
        create_environment(name, channels, packages, pip, jobs, dag, ask_for_confirmation, offline, force_download,
                           debug)
    elif command == "install":
        install_packages(name, channels, packages, pip, jobs, dag, ask_for_confirmation, offline, force_download, debug)


def create_environment(name: str, channels: list, packages: set, pip: list, jobs: int = 1, dag: bool = False,
                       ask_for_confirmation: bool = False, offline: bool = False, force_download: bool = False,
                       debug: bool = False):
    all_nodes, root = create_dag(channels, packages, [], jobs, offline, force_download, debug)
    to_install = nodes_to_install(all_nodes, root, channels)
    if dag:
        print_dag(packages, all_nodes, to_install)
    else:
        if _install_prompt(to_install, None, ask_for_confirmation):
            command = "@EXPLICIT\n"
            command += '\n'.join([f'{n.p["channel"]}/{n.p["fn"]}' for n in to_install])
            ps = run(f'bash -i -c "conda create --name {name} --file /dev/stdin"', input=command.encode(), shell=True)
    return to_install


def install_packages(name: str, channels: list, packages: set, pip: list, jobs: int = 1, dag: bool = False,
                     ask_for_confirmation: bool = False, offline: bool = False, force_download: bool = False,
                     debug: bool = False):
    local_repodata = RepositoryData.from_environment(name)
    all_nodes, root = create_dag(channels, packages, [local_repodata], jobs, offline, force_download, debug)
    to_install = nodes_to_install(all_nodes, root, channels)
    if dag:
        print_dag(packages, all_nodes, to_install, installed=local_repodata.d)
    else:
        to_install = [n for n in to_install if "installed" not in n.p]
        installed = local_repodata.d
        if _install_prompt(to_install, installed, ask_for_confirmation):
            command = "@EXPLICIT\n"
            command += '\n'.join([f'{n.p["channel"]}/{n.p["fn"]}' for n in to_install])
            ps = run('bash -i -c "conda install --file /dev/stdin"', input=command.encode(), shell=True)
    return to_install


def _get_condarc_channels() -> List[str]:
    condarc_path = os.path.expanduser(os.path.join('~', '.condarc'))
    if os.path.exists(condarc_path):
        with open(os.path.expanduser(os.path.join('~', '.condarc')), 'rt') as reader:
            channels = safe_load(reader)
            return channels['channels']
    else:
        return []


def _read_environment_description(yml: str) -> Tuple[str, list, set, set]:
    with open(yml, 'rt') as reader:
        data = safe_load(reader)
        name = data.get('name', None)
        channels = data.get('channels', [])
        package_specs = {p for p in data.get('dependencies', []) if isinstance(p, str)}
        pip_specs = [p['pip'] for p in data.get('dependencies', []) if isinstance(p, dict) and 'pip' in p]
        pip_specs = {item for sublist in pip_specs for item in sublist}
        return name, channels, package_specs, pip_specs



import bz2
import time
import copy
import json
import os
import pickle
from collections import defaultdict
from datetime import datetime, timezone
from os.path import getmtime, basename
from typing import DefaultDict, List
from urllib.error import URLError
from urllib.request import Request, urlopen


def get_repository_data(channels: list,
                        n_jobs: int = 8,
                        offline: bool = False,
                        force_download: bool = False) -> List['RepositoryData']:
    # retrieve repodata asynchronously
    urls = []
    for channel in channels:
        for arch in ['linux-64', 'noarch']:
            if channel != "defaults":
                urls.append((f"https://conda.anaconda.org/{channel}/{arch}", f"{channel}/{arch}"))
            else:
                for c in ['main', 'free', 'pro', 'r']:
                    urls.append((f"https://repo.anaconda.com/pkgs/{c}/{arch}", f"{c}/{arch}"))

    from joblib import Parallel, delayed

    return Parallel(n_jobs=n_jobs, prefer="threads")(
        delayed(download_repo)(url, name, offline, force_download) for url, name in urls)


def merge_repository_data(repository_data_list: List['RepositoryData']) -> 'RepositoryData':
    # ret = RepositoryData.empty()
    ret = repository_data_list[0]
    # deepcopy since we do not want to modify the contents of `repository_data_list[0].d`!
    # ret.d = copy.deepcopy(repository_data_list[0].d)
    for r in repository_data_list[1:]:
        for name, versions in r.d.items():
            for version in versions:
                if version not in ret.d[name]:
                    ret.d[name][version] = r.d[name][version]
    return ret


def parse_repository_data(repository_name: str, json_data: dict) -> DefaultDict[str, dict]:
    ret = defaultdict(dict)

    if json_data is None:
        return ret

    packages = json_data["packages"]

    for filename, p in packages.items():
        new_build = p["build"].rsplit("_", 1)[0]  # remove _buildnumber from build
        new_version = (p["version"], new_build)
        build_number = p["build_number"]

        p["fn"] = filename

        # different urls for main repositories
        if repository_name.split("/")[0] in ["main", "free", "pro", "r"]:
            p["channel"] = f"https://repo.continuum.io/pkgs/{repository_name}"
        else:
            p["channel"] = f"https://conda.anaconda.org/{repository_name}"

        if new_version not in ret[p["name"]] or ret[p["name"]][new_version]["build_number"] < build_number:
            ret[p["name"]][new_version] = p

    for name, versions in ret.items():
        for v_name, v_value in versions.items():
            v_value["reponame"] = repository_name
    return ret


class RepositoryData:
    def __init__(self, name: str, data: defaultdict):
        self.name = name
        self.d = data

    @classmethod
    def empty(cls) -> 'RepositoryData':
        return cls(None, defaultdict(dict))

    @classmethod
    def _from_json(cls,
                   name: str,
                   json_data: dict) -> 'RepositoryData':
        repository_data = parse_repository_data(name, json_data)
        return cls(name, repository_data)

    @classmethod
    def _from_packed(cls, name: str, file: str, method=pickle) -> 'RepositoryData':
        with open(file, 'rt' if method in {json} else 'rb') as reader:
            repository_data = method.load(reader)
        return cls(name, repository_data)

    @classmethod
    def _from_file(cls, name: str, filename: str, **kwargs) -> 'RepositoryData':
        opener = bz2.open if filename.endswith('.bz2') else open
        with opener(filename, 'rb') as reader:
            json_data = json.load(reader)
        return cls._from_json(name, json_data, **kwargs)

    @classmethod
    def from_environment(cls, environment: str) -> 'RepositoryData':
        ret = defaultdict(dict)
        # This is a linux only, default path only version of conda's `list_all_known_prefixes`.
        environments_file = os.path.expanduser('~/.conda/environments.txt')
        if not os.path.exists(environments_file):
            raise ValueError("Could not read environment information from ~/.conda/environments.txt.")
        with open(environments_file, 'rt') as reader:
            # FIXME assumption that first line of environments.txt corresponds to base env path might be incorrect
            default_env_path = reader.readline().strip()
            env_map = {basename(path): path for path in map(str.strip, reader)}
            # env_map[default_env_name] = default_env_path
        if environment == "base" or environment == "":
            meta_dir = os.path.join(default_env_path, "conda-meta")
        else:
            env_path = env_map.get(environment)
            if env_path is None:
                raise ValueError(f"No such environment {environment}.")
            meta_dir = os.path.join(env_path, "conda-meta")

        for file in os.listdir(meta_dir):
            if not file.endswith(".json"):
                continue

            with open(os.path.join(meta_dir, file), 'rb') as reader:
                p = json.load(reader)
                name = p["name"]
                p["installed"] = True
            new_version = (p["version"], p["build"])
            ret[name][new_version] = p
        return cls(environment, ret)

    @classmethod
    def from_url(cls,
                 url: str,
                 repository_name: str,
                 offline: bool = False,
                 force_download: bool = False) -> ('RepositoryData', float):
        repodata_url = url + "/repodata.json.bz2"
        local_dir = os.path.expanduser(os.path.join(_cachedir(), 'conda', 'repos', repository_name))
        local_file = os.path.join(local_dir, 'repodata.dump')
        headers = {}

        if offline:  # TODO if offline check if pickle exist
            return cls._from_packed(repository_name, local_file), None

        if not force_download and os.path.exists(local_file):
            local_timestamp = _mtime(local_file)
            d = local_timestamp.strftime('%a, %d %b %Y %H:%M:%S %Z')
            headers["If-Modified-Since"] = d
        request = Request(repodata_url, headers=headers)
        # The following code is slightly unwieldy, sorry for that
        with urlopen(Request(repodata_url, method='HEAD')) as conn:
            last_modified = conn.headers.get('last-modified', None)
            if last_modified is not None:
                last_modified = datetime.strptime(last_modified, '%a, %d %b %Y %H:%M:%S %Z')
                last_modified = last_modified.replace(tzinfo=timezone.utc).timestamp()
                if os.path.exists(local_file):
                    if last_modified <= _mtime_timestamp(local_file) and not force_download:
                        return cls._from_packed(repository_name, local_file), last_modified
        try:
            with urlopen(request) as conn:
                json_data_compressed = conn.read()
                data = json.loads(bz2.decompress(json_data_compressed))
                repodata = cls._from_json(repository_name, data)
                return repodata, last_modified
        except URLError as e:
            if e.getcode() == 304:  # NOT MODIFIED
                return cls._from_packed(repository_name, local_file), last_modified
            else:
                raise e


def download_repo(url: str, name: str, offline: bool = False, force_download: bool = False):
    repodata, remote_mtime = RepositoryData.from_url(url, name, offline, force_download)
    if not offline or force_download:
        _store_repodata(repodata, remote_mtime=remote_mtime, force_store=force_download)
    return repodata


def _store_repodata(repodata: 'RepositoryData', force_store=False, remote_mtime=None, method=pickle):
    name = repodata.name
    local_dir = os.path.expanduser(os.path.join(_cachedir(), 'conda', 'repos', name))
    local_file = os.path.join(local_dir, 'repodata.dump')
    if os.path.exists(local_file):
        local_mtime = _mtime_timestamp(local_file)
        if force_store or not remote_mtime or remote_mtime > local_mtime:
            # print(f"Remote {name}/repodata is newer, updating {local_file}.")
            with open(local_file, 'wt' if method in {json} else 'wb') as writer:
                method.dump(repodata.d, writer)
            os.utime(local_file, (remote_mtime, remote_mtime))
        else:
            # print(f"{local_file} is up to date")
            pass
    else:
        # print(f"Saving {name}/repodata to {local_file}")
        os.makedirs(local_dir, exist_ok=True)
        with open(local_file, 'wt' if method in {json} else 'wb') as writer:
            method.dump(repodata.d, writer)
        if remote_mtime:
            os.utime(local_file, (remote_mtime, remote_mtime))


def _mtime(filename: str) -> datetime:
    return datetime.fromtimestamp(getmtime(filename))


def _mtime_timestamp(filename: str) -> float:
    return _mtime(filename).replace(tzinfo=timezone.utc).timestamp()


def _cachedir() -> str:
    return os.getenv("XDG_CACHE_HOME", os.path.expanduser("~/.cache"))



from mungo.conda.models.version import VersionOrder

mungoversion = "0.5.4"


def vparse(x: str, add: str = "") -> VersionOrder:
    # TODO: remove strip?
    if x != "*":
        x = x.strip("*").strip(".")
    return VersionOrder(f"{x}{add}")


def print_dag(packages: list, all_nodes: list, to_install: list, installed: dict = None):
    from graphviz import Digraph
    dot = Digraph()
    # dot.attr("graph", splines="false")

    # ugly, but sufficient for the time being
    to_install2 = {node.p['name']: node.p for node in to_install}
    installed_keys = installed.keys() if installed is not None else {}
    changed = installed_keys & to_install2.keys()
    local_versions = {name: list(installed[name].values())[0] for name in changed}
    remote_versions = {name: data for name, data in to_install2.items() if name in changed}
    version_changes = {package: (local_versions[package], remote_versions[package]) for package in changed}
    upgrades = {package: (p_from, p_to)
                for (package, (p_from, p_to)) in version_changes.items()
                # if vparse(([p_from['version'], p_from['build']])) < vparse(([p_to['version'], p_to['build']]))
                if vparse(p_from['version']) < vparse(p_to['version'])
                }
    downgrades = {package: (p_from, p_to)
                  for (package, (p_from, p_to)) in version_changes.items()
                  if vparse(p_from['version']) > vparse(p_to['version'])
                  # if vparse(([p_from['version'], p_from['build']])) > vparse(([p_to['version'], p_to['build']]))}
                  }
    # nodes
    out_nodes = set()
    for v in to_install:
        # color = "white" if v.name not in changed else "red"
        requested = v.name in packages
        upgrade = v.name in upgrades
        downgrade = v.name in downgrades
        local = v.name in installed_keys
        kwargs = {
            # 'shape': "house" if upgrade else "invhouse" if downgrade else "folder" if local else "note",
            'shape': "folder" if local else "note",
            'fillcolor': "#C0E2E7" if requested else "#CFE7C0" if upgrade else "#E7C6C0" if downgrade else "white",
            'style': 'filled',
            'fontname': 'Fira Sans',
        }
        dot.node(f"{v.name}_{v.version}", f"{v.name}\n{v.version[0]} {v.version[1]}", **kwargs)
        out_nodes.add(v)

    # edges
    for v in to_install:
        for c in v.out_nodes:
            if c in out_nodes:
                for d in v.p["depends"]:
                    if d.startswith(c.name):  # finde dependency
                        kwargs = {
                            'fontname': 'Fira Sans',
                        }
                        dot.edge(f"{v.name}_{v.version}", f"{c.name}_{c.version}",
                                 label='\n'.join(d.split(' ')),
                                 weight=f"{v.distance_root}",
                                 **kwargs)
    dot.render('graph.tmp', view=True)
    print(dot.source)



import sys
import unittest
from unittest.mock import patch

from mungo.cli import main


class TestPackageSplitting(unittest.TestCase):

    def test_profile_install(self):
        with patch.object(sys, 'argv', ["mungo", "install", "keras", "python>=3.7", "--offline", "--dag"]):
            print(sys.argv)
            main()


if __name__ == '__main__':
    unittest.main()



import os
import unittest
from subprocess import run


class TestPackageSplitting(unittest.TestCase):

    def setUp(self):
        shlvl = int(os.getenv('CONDA_SHLVL', '0'))
        self.envs = []
        while shlvl > 0:
            self.envs.append(os.getenv('CONDA_DEFAULT_ENV'))
            # TODO proper way to call conda commands
            run('bash -i -c "conda deactivate"', shell=True)
            shlvl = int(os.getenv('CONDA_SHLVL', '0'))

    def tearDown(self):
        for env in self.envs:
            run(f'bash -i -c "conda activate {env}"', shell=True)

    def test_get_base_prefix(self):
        run('bash -i -c "conda activate"', shell=True)

        shlvl = int(os.getenv('CONDA_SHLVL', '0'))
        self.assertEqual(shlvl, 0)

        prefix = os.getenv('CONDA_PREFIX')
        # assert: prefix == base environment *path*

        default_env = os.getenv('CONDA_DEFAULT_ENV')
        self.assertEqual(default_env, 'base')

    def test_get_active_prefix(self):
        raise NotImplementedError()

    def test_get_installed_packages(self):
        raise NotImplementedError()

    def test_get_channels(self):
        raise NotImplementedError()


if __name__ == '__main__':
    unittest.main()



import unittest

from mungo.solver import split_package_constraint


class TestPackageSplitting(unittest.TestCase):

    # def test_build(self):
    #     split = split_package_constraint("pysocks 4 rc3")
    #     self.assertEqual(split, ("pysocks", [['4', 'rc3']]))
    #     # TODO what's the expected behaviour?

    def test_multisplit(self):
        split = split_package_constraint("pysocks >=1.5.6,<2.0,!=1.5.7")
        self.assertEqual(split, ("pysocks", [[">=", "1.5.6"], ["<", "2.0"], ["!=", "1.5.7"]]))

    def test_multisplit_whitespace(self):
        split = split_package_constraint("pysocks>=   1.5.6,  <  2.0,!= 1.5.7")
        self.assertEqual(split, ("pysocks", [[">=", "1.5.6"], ["<", "2.0"], ["!=", "1.5.7"]]))

    def test_unknown(self):
        split = split_package_constraint("pysocks 1=4")
        # TODO what's the expected behaviour?, assertion


if __name__ == '__main__':
    unittest.main()



import unittest

from mungo.cli import create_environment
from mungo.repositorydata import RepositoryData


def packages(nodes):
    return set(map(lambda x: (x.name, x.version), nodes))


# TODO: tear down, do not print and open DAG
class TestDependencies(unittest.TestCase):

    def setUp(self):
        repodata = RepositoryData._from_file("fakerepo/linux-64", "fakerepo/linux-64/repodata.json",
                                             cache=True,
                                             remote_mtime=0)
        repodata = RepositoryData._from_file("fakerepo/noarch", "fakerepo/noarch/repodata.json",
                                             cache=True,
                                             remote_mtime=0)

    def test_simple_dependency_BA(self):
        p = packages(create_environment("foo", ["fakerepo"], {"B"},
                                        [], jobs=1, dag=True, ask_for_confirmation=True, offline=True,
                                        force_download=False))
        self.assertEqual(p, {("A", ("0.1.0", "py37")), ("B", ("0.3.0", "py37"))})

    def test_simple_dependency_CAB(self):
        p = packages(create_environment("foo", ["fakerepo"], {"C"},
                                        [], jobs=1, dag=True, ask_for_confirmation=True, offline=True,
                                        force_download=False))
        self.assertEqual(p, {("C", ("0.2.0", "py37")), ("B", ("0.2.0", "py37")), ("A", ("0.1.0", "py37"))})

    def test_or_dependency_DB(self):
        p = packages(create_environment("foo", ["fakerepo"], {"D"},
                                        [], jobs=1, dag=True, ask_for_confirmation=True, offline=True,
                                        force_download=False))
        self.assertEqual(p, {("D", ("0.1.0", "py37")), ("B", ("0.3.0", "py37")), ("A", ("0.1.0", "py37"))})

    def test_buildname_dependency_FG(self):
        p = packages(create_environment("foo", ["fakerepo"], {"F"},
                                        [], jobs=1, dag=True, ask_for_confirmation=True, offline=True,
                                        force_download=False))
        self.assertEqual(p, {("F", ("0.1.0", "py37")), ("G", ("0.1.0", "foo"))})

    def test_cyclic_dependency_HI(self):
        p = packages(create_environment("foo", ["fakerepo"], {"H"},
                                        [], jobs=1, dag=True, ask_for_confirmation=True, offline=True,
                                        force_download=False))
        self.assertEqual(p, {("H", ("0.1.0", "py37")), ("I", ("0.1.0", "py37"))})


if __name__ == '__main__':
    unittest.main()



import setuptools
from mungo.common import mungoversion

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="mungo",
    version=f"{mungoversion}",
    author="Christopher Schröder, Till Hartmann",
    author_email="christopher.schroeder@uni-due.de, till.hartmann@udo.edu",
    description="Quick environment solver",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/christopher-schroeder/mungo",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: Unix",
    ],
    entry_points={
        'console_scripts': ['mungo=mungo.cli:main'],
    },
    install_requires=['pulp', 'pyyaml', 'joblib', 'graphviz'],
)



import re
from collections import defaultdict, namedtuple
from functools import lru_cache
from functools import partial
from typing import Dict, Callable, List, Tuple, DefaultDict, Set

from pulp import *
from pulp import LpProblem, LpMinimize, LpStatusOptimal, LpStatus

from mungo.common import vparse
from mungo.repositorydata import get_repository_data, merge_repository_data, RepositoryData

from time import time

Version = namedtuple("Version", "version build")


def equal_compare(y):
    p = re.compile(y.replace(".*", "*").replace("*", "(\D.*)??\Z"))
    return lambda x: bool(p.match(x.version))


# 3.*
# 3a
# 3*
# 3.0a

operator_function: Dict[str, Callable[[Version, Version], bool]] = {
    ">=": lambda x, y: vparse(x.version) >= vparse(y.version),
    ">": lambda x, y: vparse(x.version) > vparse(y.version),
    "<=": lambda x, y: vparse(x.version) <= vparse(y.version),
    "<": lambda x, y: vparse(x.version) < vparse(y.version),
    "!=": lambda x, y: ~(vparse(x.version) == vparse(y.version))
                       and (vparse(x.version) < vparse(y.version, ".99999999999999")),
    "==": lambda x, y: vparse(x.version) == vparse(y.version),
    #    "=": lambda x, y: (vparse(x.version) >= vparse(y.version))
    #                      and (vparse(x.version) < vparse(y.version, ".99999999999999")),
}


def build_comp(x: Version, constraint: Version) -> bool:
    xb, yb = x.build, constraint.build
    return xb == yb \
           or yb.endswith("*") and xb.startswith(yb.rstrip("*")) \
           or yb == "*"


class Node:
    def __init__(self, name: str, version, p):
        self.name = name
        self.version = version
        self.distance_root = -1
        self._in_nodes = []
        self._out_nodes = defaultdict(list)
        self._x = None
        self.p = p
        self._mass_low = None
        self._mass_high = None
        self.invalid = False

    def add_in_node(self, n: 'Node'):
        self._in_nodes.append(n)

    def remove_in_node(self, n: 'Node'):
        self._in_nodes.remove(n)

    def add_out_node(self, n: 'Node'):
        self._out_nodes[n.name].append(n)

    def remove_out_node(self, n: 'Node'):
        self._out_nodes[n.name].remove(n)

    @property
    def channel_name(self) -> str:
        if self.p is None or "channel" not in self.p:
            return ""
        return self.p['channel'].split('/')[-2]

    @property
    def installed(self) -> bool:
        if self.p is None:
            return False
        # print(self.name, self.version, "installed" in self.p)
        return "installed" in self.p and self.p["installed"]

    @property
    def in_nodes(self) -> List['Node']:
        return self._in_nodes

    @property
    def out_nodes(self) -> List['Node']:
        for nodes in self._out_nodes.values():
            for n in nodes:
                yield n

    @property
    def x(self) -> LpVariable:
        # create LP variable, if not exist
        if self._x is None:
            self._x = LpVariable("{name}-{version}".format(name=self.name, version=self.version), 0, 1, LpInteger)
        return self._x

    def delete(self):
        for n in self.in_nodes:
            n.remove_out_node(self)
        for n in self.out_nodes:
            n.remove_in_node(self)


NodeDict = DefaultDict[str, Dict[str, Node]]


@lru_cache()
def valid_packages(ls: list, repodata: dict):
    # print(ls)
    all_valids = []
    for s in ls:
        name, is_valid = _valid_packages_function(s)
        all_valids.append(is_valid)

    # all_valid = lambda x: reduce((lambda a, b: a(x) and b(x)), all_valids)
    all_valid = lambda x: any([f(x) for f in all_valids])  # TODO: any or all?
    ret = [(name, x) for x in repodata.d[name] if all_valid(Version(*x))]
    return ret


def split_package_constraint(s: str) -> Tuple[str, List[List[List[str]]]]:
    or_split = s.split("|")
    ret = []
    name = None

    for o in or_split:
        term = []
        for x in o.split(","):
            res = re.split('(==|!=|>=|<=|=|<|>| )', x)
            res = [y for y in res if y != " " and len(y) > 0]
            if name is None:
                name = res.pop(0)
            term.append(res)
        ret.append(term)
    return name, ret  # foo, [[]]


# print(split_package_constain("pysocks 4 rc3"))
# print(split_package_constain("pysocks >=1.5.6,<2.0,!=1.5.7"))
# print(split_package_constain("pysocks>=   1.5.6,  <  2.0,!= 1.5.7"))
# print(split_package_constain("pysocks 1=4"))
# f()

def _valid_packages_function(s: str) -> Tuple[str, Callable[[Version], bool]]:
    name, or_constraints = split_package_constraint(s)

    and_valids = []

    for constraints in or_constraints:
        is_valids = []  # lambda x: True

        for c in constraints:
            len_c = len(c)
            if len_c == 0:
                # no version constraint
                pass
            elif len_c == 1:
                # version constraint without operator (equals ==)
                version = c[0]
                is_valids.append(equal_compare(y=version))
            elif len_c == 2:
                # version constraint
                operator, v = c
                version = Version(v, "*")
                if operator == "=":
                    is_valids.append(equal_compare(y=v))
                elif operator in [">=", ">", "<=", "<", "!=", "=="]:
                    is_valids.append(partial(operator_function[operator], y=version))
                else:
                    # version build constraint with no operator
                    version = Version(*c)
                    if version.version != "*":
                        is_valids.append(equal_compare(y=version.version))
                    is_valids.append(partial(build_comp, constraint=version))
            elif len_c == 3:
                # operator version build
                operator, v, b = c
                version = Version(v, b)
                if operator == "=":
                    is_valids.append(equal_compare(y=v))
                elif operator in [">=", ">", "<=", "<", "!=", "=="]:
                    is_valids.append(partial(operator_function[operator], y=version))
                is_valids.append(partial(build_comp, constraint=version))
            else:
                raise Exception("invalid version constraint", s)

        def all_valids(x, is_valids=is_valids):
            return all([f(x) for f in is_valids])

        and_valids.append(all_valids)
    or_valid = lambda x: any([f(x) for f in and_valids])
    return name, or_valid


def create_node(name: str,
                version: str,
                repodata: RepositoryData,
                sender: Node = None,
                all_nodes: NodeDict = defaultdict(dict),
                override_dependencies: List[str] = None,
                debug: bool = False):
    if version in all_nodes[name]:
        node = all_nodes[name][version]
        node.add_in_node(sender)  # add the parent to in-nodes
        return node

    # create new node
    if version in repodata.d[name]:
        p = repodata.d[name][version]
    else:
        p = None

    new_node = Node(name, version, p)

    if sender is not None:
        new_node.add_in_node(sender)

    all_nodes[name][version] = new_node

    # get package information from repodata
    if override_dependencies is not None:
        dependencies = override_dependencies
    else:
        dependencies = p["depends"]

    # multiple constraints may refer the same package
    # group them

    grouped_dependencies = defaultdict(list)
    for dependency in dependencies:
        name = split_package_constraint(dependency)[0]
        grouped_dependencies[name].append(dependency)
    # print("python >=3.7,<3.8.0a0", valid_packages(("python >=3.7,<3.8.0a0", ), repodata))
    # exit()

    # print("openssl 1.0.1", list(valid_packages(["openssl 1.0.1"], repodata)))
    # print("valid_packages", valid_packages(("jpeg 9b",), repodata))
    # exit()

    # print("valid_packages", valid_packages(("zlib 1.2.*",), repodata))
    # exit()

    # print("valid_packages", valid_packages(("blas 1.1 openblas",), repodata))
    # exit()

    # print("valid_packages", valid_packages(("openmpi",), repodata))
    # exit()

    # print(name, version, grouped_dependencies)

    for dependency in grouped_dependencies.values():
        d_node = None
        # make sure s is list like if string
        if isinstance(dependency, str):
            dependency = (dependency,)
        for d_name, d_version in valid_packages(tuple(dependency), repodata):
            d_node = create_node(d_name, d_version, repodata, new_node, all_nodes, debug=debug)
            new_node.add_out_node(d_node)

        if d_node is None:  # no versions were found for a dependency
            print_missing_package(dependency, debug=debug)
            new_node.invalid = True
            # raise Exception("no packages found for", *dependency)

    return new_node


def reduce_dag(all_nodes: NodeDict, channels: List[str]):
    # reduce DAG

    # for name, versions in all_nodes.items():
    #    for current_node in versions.values():
    #        print(current_node.name, current_node.version)

    channel_order = dict()
    for priority, channel in enumerate(reversed(channels)):
        if channel == "defaults":
            channel_order['main'] = priority
            channel_order['free'] = priority
            channel_order['pro'] = priority
            channel_order['r'] = priority
        else:
            channel_order[channel] = priority
        channel_order[""] = priority + 1  # highest priority for local installed packages

    nodes_to_remove = set()

    for name, versions in all_nodes.items():
        if name == "root":
            continue
        seen_keys = set()

        # if name == "ca-certificates":
        #     print(list(sorted(versions.keys(), key=lambda v: (channel_order[versions[v].channel_name] + versions[v].is_installed * (priority + 1), vparse(v)), reverse=True)))
        #     exit()

        # TODO the following line is also used in "base" and should be some kind of function import
        for v in sorted(versions.keys(),
                        key=lambda x: (channel_order[versions[x].channel_name]
                                       + versions[x].installed * (priority + 1), vparse(x[0])),
                        reverse=True):
            current_node = versions[v]
            in_node_key = tuple(sorted([(n.name, n.version) for n in current_node.in_nodes]))
            out_node_key = tuple(sorted([(n.name, n.version) for n in current_node.out_nodes]))
            key = (in_node_key, out_node_key)

            # if current_node.version[0] == "3.6.8":
            # print(current_node.channel_name, current_node.is_installed, current_node.name, current_node.version, key in seen_keys)

            if key in seen_keys:
                # the in-node out-node combination exist in a higher version of the same package
                # thus the higher version is ALWAYS preferable and this version can be removed
                nodes_to_remove.add(current_node)
            else:
                seen_keys.add(key)

    # remove the identified nodes
    for n in nodes_to_remove:
        n.delete()
        del (all_nodes[n.name][n.version])


def distances_dag(root: Node):
    q = [root]
    root.distance_root = 0

    while len(q) > 0:
        current_node = q.pop(0)
        for child in current_node.out_nodes:
            if child.distance_root == -1:
                child.distance_root = current_node.distance_root + 1
                q.append(child)


def tprint(msg):
    print(msg, "...")
    return time()


def create_dag(channels: List[str],
               packages: Set[str],
               local_repodata: List[RepositoryData],
               njobs: int = 8,
               offline: bool = False,
               force_download: bool = False,
               debug: bool = False) -> Tuple[NodeDict, Node]:

    start = tprint("Retrieving repository data ...")
    repodata_chunks = get_repository_data(channels, njobs, offline, force_download)
    print(f"Done ({time() - start:0.2f}s)")

    start = tprint("Merging repository data ...")
    repodata = merge_repository_data(local_repodata + repodata_chunks)
    print(f"Done ({time() - start:0.2f}s)")

    # add installed packages to the repodata

    # create dag
    start = tprint("Creating dag ...")
    default_packages = set()
    # default_packages = {"sqlite", "wheel", "pip"}
    all_nodes = defaultdict(dict)
    root = create_node("root", ("0", "0"), repodata, all_nodes=all_nodes,
                       override_dependencies=list(packages | default_packages), debug=debug)
    print(f"Done ({time() - start:0.2f}s)")

    # reduce DAG
    start = tprint("Reduce dag ...")
    reduce_dag(all_nodes, channels)
    print(f"Done ({time() - start:0.2f}s)")

    distances_dag(root)
    return all_nodes, root


def nodes_to_install(all_nodes: NodeDict, root: Node, channels: List[str]) -> List[Node]:
    channel_order = dict()
    for priority, channel in enumerate(reversed(channels)):
        if channel == "defaults":
            channel_order['main'] = priority
            channel_order['free'] = priority
            channel_order['pro'] = priority
            channel_order['r'] = priority
        else:
            channel_order[channel] = priority
        channel_order[""] = priority + 1  # highest priority for local installed packages

    # create ILP problem
    prob = LpProblem("DependencySolve", LpMinimize)

    # build LP on the reduced DAG
    objective = []
    prob += root.x == 1

    # n.p['channel'].split('/')[-2]

    # create brother information
    for name, versions in all_nodes.items():
        big_brother = None
        # small hack to make local always be the highest

        for v in sorted(versions.keys(),
                        key=lambda v: (
                                channel_order[versions[v].channel_name] + versions[v].installed * (priority + 1),
                                vparse(v[0])),
                        reverse=True):
            n = versions[v]
            n.big_brother = big_brother
            big_brother = n


    def mass(n: Node):
        if n is None:
            return 0, 0

#        if n.name == "attrs" and n.version == ('15.2.0', 'py36'):
#            for children in n._out_nodes.values():
#                print([(c.name, c.version) for c in children])
#        print("\tmass", n.name, n.version)

        if n._mass_low is None:
            # n._mass_low = 0  # temporarily set mass to 0 for loops in "DAG"
            # n._mass_high = 0  # temporarily set mass to 0 for loops in "DAG"
            n._mass_low = 1
            n._mass_high = 1
            m_low = 0
            m_high = 0
            # print(n.name, n.distance_root)
            for children in n._out_nodes.values():
                children_masses = [mass(c) for c in children if c.distance_root > n.distance_root]
                if len(children_masses) > 0:
                    m_low += min([c for c, _ in children_masses])
                    m_high += max([c for _, c in children_masses])
                else:
                    m_low = 0
                    m_high = 1
            # n._mass_low = m_low + 1 + mass(n.big_brother)[1]
            # n._mass_high = m_high + 1 + mass(n.big_brother)[1]
            n._mass_high = m_high - m_low + 1 + mass(n.big_brother)[1]
            n._mass_low = 1 + mass(n.big_brother)[1]
        return n._mass_low, n._mass_high

    # for name, versions in all_nodes.items():
    #     for version, n in versions.items():
    #         print(name, version, mass(n))

    #breadth first search
    seen = [root]
    current_index = 0
    while current_index < len(seen):
        e = seen[current_index]
        for c in e.out_nodes:
            if c not in seen:
                seen.append(c)
        current_index += 1

#    for n in reversed(seen):
#        print(n.name, n.version)
#        mass(n)
#    exit()

    # create ILP
    constraints = []
    for name, versions in all_nodes.items():
        # constraint: install at most one version of a package
        constraints.append(pulp.lpSum(n.x for n in versions.values()) <= 1)
        # exclude invalid node
        # constraint: if a parent is installed, one version of each dependency must be installed too
        for current_node in versions.values():
            #            print(current_node.name, current_node.version)

            if current_node.invalid:  # exlude invalid nodes
                constraints.append(current_node.x == 0)
                continue

            for out_group in current_node._out_nodes.values():
                constraints.append(pulp.lpSum([n.x for n in out_group]) >= current_node.x)

        # storing the objectives
        objective.extend([mass(n)[0] * n.x for n in versions.values()])
        # print([(n.name, n.normalized_version, n.factor) for n in versions.values()])

        if current_node is root:  # no no_install for root  # FIXME current_node might be uninitialized
            continue

    prob.setObjective(pulp.lpSum(objective))
    prob.extend(constraints)
    # prob.writeLP("WhiskasModel.lp")
    prob.solve()

    if prob.status != LpStatusOptimal:
        print(f"ERROR: Solution is not optimal (status: {LpStatus[prob.status]}).\nAborting.", file=sys.stderr)
        exit(1)

    # for v in prob.variables():
    #     print(v.name, v.varValue)

    # collect all install nodes
    install_nodes = []
    for _, versions in sorted(all_nodes.items()):
        for n in versions.values():
            # print(n.name, n.version, n.x.varValue, "installed" in n.p)
            if n.p is None:  # skip root
                continue
            x = n.x
            if x.varValue == 1.0:
                install_nodes.append(n)

    return install_nodes


missing_dependencies = set()


def print_missing_package(dependency: list, debug=False):
    if not debug:
        return
    # print missing package information for a dependency constraint
    # only print once by remembering
    dependency = tuple(dependency)
    if dependency not in missing_dependencies:
        print("WARNING:", "no packages found for", *dependency, file=sys.stderr)
        missing_dependencies.add(dependency)
